# Copyright (c) 2010 Infrae / Technical University Delft. All rights reserved.
# See also LICENSE.txt

import sys
import os
import commands

FEDORA_INSTALL_PROPERTIES="""
ri.enabled=true
messaging.enabled=false
apia.auth.required=false
database.jdbcDriverClass=org.apache.derby.jdbc.EmbeddedDriver
ssl.available=false
database.jdbcURL=jdbc\:derby\:%(path)s/derby/fedora3;create\=true
database.password=fedoraAdmin
database.username=fedroaAdmin
tomcat.shutdown.port=8005
deploy.local.services=false
xacml.enabled=false
tomcat.http.port=%(port)s
fedora.serverHost=%(host)s
database=included
database.driver=included
fedora.serverContext=fedora
tomcat.home=%(path)s/tomcat
fedora.home=%(path)s
install.type=custom
fesl.enabled=false
servlet.engine=included
fedora.admin.pass=%(passwd)s"""

FEDORA_HOST = 'localhost'
FEDORA_PORT = '8080'
FEDORA_PASSWD = 'fedoraAdmin'

def install_fedora():
    jarfile = os.path.join(os.getcwd(), 'parts', 'java-dependencies',
                           'fcrepo-installer-3.3.jar')

    if not os.path.isfile(jarfile):
        print >> sys.stderr, ('fcrepo-installer-3.3.jar is missing, '
                              'run buildout first')
        sys.exit(1)

    java_version = (
        commands.getoutput('java -version').splitlines() or [''])[0]
    
    if not java_version.startswith('java version "1.6'):
        print >> sys.stderr, ('can not find java, or wrong version')
        sys.exit(1)
    
    fedora_path = os.path.join(os.getcwd(), 'parts', 'fedora')
    install_props = FEDORA_INSTALL_PROPERTIES % {'host': FEDORA_HOST,
                                                 'port': FEDORA_PORT,
                                                 'passwd': FEDORA_PASSWD,
                                                 'path': fedora_path}
    fp = open('/tmp/fedora_install.properties', 'w')
    fp.write(install_props)
    fp.close()

    os.system('java -jar "%s" /tmp/fedora_install.properties' % jarfile)

def start_fedora():
    fedora_path = os.path.join(os.getcwd(), 'parts', 'fedora')
    cmd = ('export FEDORA_HOME="%s" && export CATALINA_HOME="%s/tomcat" && '
           'sh "%s/tomcat/bin/catalina.sh" run' % (fedora_path,
                                                   fedora_path,
                                                   fedora_path))
    os.system(cmd)



# Copyright (c) 2010 Infrae / Technical University Delft. All rights reserved.
# See also LICENSE.txt

import urllib

from lxml import etree

NSMAP = {'wadl': 'http://research.sun.com/wadl/2006/10'}

class WADLMethod(object):
    def __init__(self, id, name, api):
        self.id = id
        self.api = api
        self.name = name
        self.url = u'/'.join(self.api.doc.xpath(
            '//wadl:method[@id="%s"]//ancestor::wadl:resource/@path' % self.id,
            namespaces=NSMAP))
        self.url = self.url.replace('//', '/').replace(
            '%', '%%').replace('{', '%(').replace('}', ')s')

        # XXX hack to fix broken fedora wadl.
        if not self.url.startswith('/objects'):
            self.url = '/objects%s' % self.url
        
    def __call__(self, **params):
        url = self.url % params
        return WADLRequest(url, self)

class WADLRequest(object):
    def __init__(self, url, method):
        self.url = url
        self.method = method
        self.headers = self.method.api.connection.form_headers.copy()

        self.param_types = {}
        self.undocumented_params = {} # needed in searchOjbects
        self.default_values = {}
        
        for param in self.method.api.doc.xpath(
            '//wadl:method[@id="%s"]/wadl:request/wadl:param' % self.method.id,
            namespaces=NSMAP):
            name = param.attrib['name']
            param_type = {'xs:int': int,
                          'xs:boolean': bool,
                          'xs:string': unicode}[param.attrib['type']]
            self.param_types[name] = param_type
            default_value = param.attrib.get('default')
            if default_value:
                self.default_values[name] = default_value
        

    def submit(self, body='', **params):
        qs = self.default_values.copy()
        for param, value in params.items():
            param_type = self.param_types.get(param)
            if param_type is None:
                raise KeyError('Method "%s" has no param "%s"' % (
                    self.method.id, param))
            if not isinstance(value, param_type):
                raise TypeError(
             'Expected %s for param "%s" on method "%s", got %s instead' % (
                    param_type,
                    param,
                    self.method.id,
                    value.__class__))
            
            if param_type is bool:
                value = unicode(param_type).lower()
            value = unicode(value)
            qs[param] = value
        for param, value in self.undocumented_params.items():
            if not param in qs:
                qs[param] = value
        if qs:
            self.url = '%s?%s' % (self.url, urllib.urlencode(qs))

        return self.method.api.connection.open(self.url,
                                               body,
                                               self.headers,
                                               method=self.method.name)
    
class API(object):
    def __init__(self, connection):
        self.connection = connection
        fp = self.connection.open('/objects/application.wadl')
        wadl_xml = fp.read()
        fp.close()
        self.doc = etree.fromstring(wadl_xml)
        for method in self.doc.xpath('//wadl:method',
                                     namespaces=NSMAP):
            method_id = method.attrib['id']
            method_name = method.attrib['name']
            self.__dict__[method_id] = WADLMethod(method_id, method_name, self)
        



# Copyright (c) 2010 Infrae / Technical University Delft. All rights reserved.
# See also LICENSE.txt

from fcrepo.datastream import FedoraDatastream, RELSEXTDatastream, DCDatastream

class FedoraObject(object):
    def __init__(self, pid, client):
        self.pid = pid
        self.client = client
        self._info = self.client.getObjectProfile(self.pid)
        self._dsids = None # load lazy
        self._methods = None
        
    def _setProperty(self, name, value):
        msg = u'Changed %s object property' % name
        kwargs = {name: value, 'logMessage': msg}
        self.client.updateObject(self.pid, **kwargs)
        self._info = self.client.getObjectProfile(self.pid)

    label = property(lambda self: self._info['label'],
                     lambda self, value: self._setProperty('label', value))
    ownerId = property(lambda self: self._info['ownerId'],
                       lambda self, value: self._setProperty('ownerId', value))
    state = property(lambda self: self._info['state'],
                           lambda self, value: self._setProperty('state',
                                                                 value))
    # read only
    createdDate = property(lambda self: self._info['createdDate'])
    lastModifiedDate = property(lambda self: self._info['lastModifiedDate'])

    def datastreams(self):
        if self._dsids is None:
            self._dsids = self.client.listDatastreams(self.pid)
        return self._dsids

    def __iter__(self):
        return iter(self.datastreams())
    
    def __in__(self, dsid):
        return dsid in self.datastreams()

    def __getitem__(self, dsid):
        if dsid == 'DC':
            return DCDatastream(dsid, self)
        elif dsid == 'RELS-EXT':
            return RELSEXTDatastream(dsid, self)
        else:
            return FedoraDatastream(dsid, self)

    def __delitem__(self, dsid):
        self.client.deleteDatastream(self.pid, dsid)
        self._dsids = None

    def delete(self, **params):
        self.client.deleteObject(self.pid, **params)
        
    def addDataStream(self, dsid, body='', **params):            
        self.client.addDatastream(self.pid, dsid, body, **params)
        self._dsids=None

    def methods(self):
        if self._methods is None:
            self._methods = self.client.getAllObjectMethods(self.pid)
        return [m[1] for m in self._methods]

    def call(self, method_name, **params):
        for sdef, method in self._methods:
            if method == method_name:
                break
        else:
            raise KeyError('No such method: %s' % method_name)
        
        return self.client.invokeSDefMethodUsingGET(self.pid, sdef,
                                                    method, **params)



import unittest
import doctest
import datetime

FLAGS = doctest.NORMALIZE_WHITESPACE + doctest.ELLIPSIS
GLOBS = {'datetime':datetime}

def test_suite():

    suite = unittest.TestSuite()
    suite.addTests([
        doctest.DocFileSuite('README.txt',
                             package='fcrepo',
                             globs=GLOBS,
                             optionflags=FLAGS),
        ])

    return suite




# Copyright (c) 2010 Infrae / Technical University Delft. All rights reserved.
# See also LICENSE.txt

import urllib

from lxml import etree
from lxml.builder import ElementMaker

from fcrepo.wadl import API
from fcrepo.utils import NS
from fcrepo.object import FedoraObject

NSMAP = {'foxml': 'info:fedora/fedora-system:def/foxml#'}

class FedoraClient(object):
    def __init__(self, connection):
        self.api = API(connection)

    def getNextPID(self, namespace, numPIDs=1, format=u'text/xml'):
        request = self.api.getNextPID()
        response = request.submit(namespace=namespace,
                                  numPIDs=numPIDs,
                                  format=format)
        xml = response.read()
        response.close()
        doc = etree.fromstring(xml)
        ids = [id.decode('utf8') for id in doc.xpath('/pidList/pid/text()')]
        if len(ids) == 1:
            return ids[0]
        return ids
        
    def createObject(self, pid, label, state=u'A'):
        foxml = ElementMaker(namespace=NSMAP['foxml'], nsmap=NSMAP)
        foxml_state = {'A': u'Active',
                       'I': u'Inactive',
                       'D': u'Deleted'}[state]
        doc = foxml.digitalObject(
            foxml.objectProperties(
              foxml.property(NAME="info:fedora/fedora-system:def/model#state",
                             VALUE=state),
              foxml.property(NAME="info:fedora/fedora-system:def/model#label",
                             VALUE=label)
              ),
            VERSION='1.1', PID=pid)
        body = etree.tostring(doc, encoding="UTF-8", xml_declaration=False)
        
        request = self.api.createObject(pid=pid)
        request.headers['Content-Type'] = 'text/xml; charset=utf-8'
        response = request.submit(body, state=state[0], label=label)
        return self.getObject(pid)
    
    def getObject(self, pid):
        return FedoraObject(pid, self)

    def getObjectProfile(self, pid):
        request = self.api.getObjectProfile(pid=pid)
        response = request.submit(format=u'text/xml')
        xml = response.read()
        response.close()
        doc = etree.fromstring(xml)
        result = {'ownerId': u''}
        for child in doc:
            # rename elementnames to match property names in foxml
            name = {'objLabel': 'label',
                    'objOwnerId': 'ownerId',
                    'objCreateDate': 'createdDate',
                    'objLastModDate': 'lastModifiedDate',
                    'objState': 'state'}.get(child.tag)
            if name is None or child.text is None:
                continue
            value = child.text
            if not isinstance(value, unicode):
                value = value.decode('utf8')
            result[name] = value
        return result

    def updateObject(self, pid, body='', **params):
        request = self.api.updateObject(pid=pid)
        response = request.submit(body, **params)

    def deleteObject(self, pid, **params):
        request = self.api.deleteObject(pid=pid)
        response = request.submit(**params)
        
    def listDatastreams(self, pid):
        request = self.api.listDatastreams(pid=pid)
        response = request.submit(format=u'text/xml')
        xml = response.read()
        response.close()
        doc = etree.fromstring(xml)
        return doc.xpath('/objectDatastreams/datastream/@dsid')

    def addDatastream(self, pid, dsid, body='', **params):
        if dsid == 'RELS-EXT' and not body:
            body = ('<rdf:RDF xmlns:rdf="%s"/>' % NS.rdf)
            params['mimeType'] = u'application/rdf+xml'
            params['formatURI'] = (
                u'info:fedora/fedora-system:FedoraRELSExt-1.0')

        if params.get('controlGroup', u'X') == u'X':
            if not 'mimeType' in params:
                params['mimeType'] = u'text/xml'
        if not 'mimeType' in params:
            params['mimeType'] = u'application/binary'
            
        if 'checksumType' not in params:
            params['checksumType'] = u'MD5'

        params = self._fix_ds_params(params)

        request = self.api.addDatastream(pid=pid, dsID=dsid)
        request.headers['Content-Type'] = params['mimeType']
        response = request.submit(body, **params)        

    def _fix_ds_params(self, params):
        for name, param in params.items():
            newname = {'label': 'dsLabel',
                       'location': 'dsLocation',
                       'state': 'dsState'}.get(name, name)
            if newname != name:
                params[newname] = param
                del params[name]
                
        return params
        
    def getDatastreamProfile(self, pid, dsid):
        request = self.api.getDatastreamProfile(pid=pid, dsID=dsid)
        response = request.submit(format=u'text/xml')
        xml = response.read()
        response.close()
        doc = etree.fromstring(xml)
        result = {}
        for child in doc:
            # rename elementnames to match property names in foxml
            name = {'dsLabel': 'label',
                    'dsVerionId': 'versionId',
                    'dsCreateDate': 'createdDate',
                    'dsState': 'state',
                    'dsMIME': 'mimeType',
                    'dsFormatURI': 'formatURI',
                    'dsControlGroup': 'controlGroup',
                    'dsSize': 'size',
                    'dsVersionable': 'versionable',
                    'dsInfoType': 'infoType',
                    'dsLocation': 'location',
                    'dsLocationType': 'locationType',
                    'dsChecksum': 'checksum',
                    'dsChecksumType': 'checksumType'}.get(child.tag)
            if name is None or child.text is None:
                continue
            value = child.text
            if not isinstance(value, unicode):
                value = value.decode('utf8')
            result[name] = value
        return result

    def modifyDatastream(self, pid, dsid, body='', **params):
        params = self._fix_ds_params(params)
        request = self.api.modifyDatastream(pid=pid, dsID=dsid)
        response = request.submit(body, **params)
        
    def getDatastream(self, pid, dsid):
        request = self.api.getDatastream(pid=pid, dsID=dsid)
        return request.submit()

    def deleteDatastream(self, pid, dsid, **params):
        request = self.api.deleteDatastream(pid=pid, dsID=dsid)
        return request.submit(**params)

    def getAllObjectMethods(self, pid, **params):
        params['format'] = u'text/xml'
        request = self.api.getAllObjectMethods(pid=pid)
        response = request.submit(**params)
        xml = response.read()
        doc = etree.fromstring(xml)
        method_list = []
        for sdef_el in doc:
            sdef = sdef_el.attrib['pid']
            for method_el in sdef_el:
                method = method_el.attrib['name']
                method_list.append((sdef, method))
        response.close()
        return method_list

    def invokeSDefMethodUsingGET(self, pid, sdef, method, **params):
        request = self.api.invokeSDefMethodUsingGET(pid=pid, sDef=sdef,
                                                    method=method)
        return request.submit(**params)

        
    def searchObjects(self, query, fields, maxResults=10):
        field_params = {}
        assert isinstance(fields, list)
        for field in fields:
            field_params[field] = u'true'
            
        token = True
        NS = 'http://www.fedora.info/definitions/1/0/types/'
        while token:
            if token is True:
                token = False
            request = self.api.searchObjects()
            request.undocumented_params = field_params

            if token:
                response = request.submit(query=query,
                                          sessionToken=token,
                                          maxResults=maxResults,
                                          resultFormat=u'text/xml')
            else:
                response = request.submit(query=query,
                                          maxResults=maxResults,
                                          resultFormat=u'text/xml')

            xml = response.read()
            response.close()
            doc = etree.fromstring(xml)
                
            tokens = doc.xpath('//f:token/text()', namespaces={'f': NS})
            if tokens:
                token = tokens[0].decode('utf8')
            else:
                token = False
            
            for result in doc.xpath('//f:objectFields', namespaces={'f': NS}):
                data = {}
                for child in result:
                    field_name = child.tag.split('}')[-1].decode('utf8')
                    value = child.text
                    if not isinstance(value, unicode):
                        value = value.decode('utf8')
                        data[field_name] = value
                yield data

                
    def searchTriples(self, query, lang='sparql', format='Sparql',
                      limit=100, type='tuples', dt='on', flush=True):
        flush = str(flush).lower()
        url = u'/risearch?%s' % urllib.urlencode({'query':query,
                                                  'lang':lang,
                                                  'flush': flush,
                                                  'format':format,
                                                  'limit':limit,
                                                  'type':type,
                                                  'dt':dt})
        headers = {'Accept:': 'text/xml'}
        response = self.api.connection.open(url, '', headers, method='POST')
        xml = response.read()
        doc = etree.fromstring(xml)
        NS = 'http://www.w3.org/2001/sw/DataAccess/rf1/result' # ouch, old!
        for result in doc.xpath('//sparql:result', namespaces={'sparql': NS}):
            data = {}
            for el in result:
                name = el.tag.split('}')[-1]
                value = {}
                uri = el.attrib.get('uri')
                if uri:
                    value['value'] = uri.decode('utf8')
                    value['type'] = 'uri'
                else:
                    value['type'] = 'literal'
                    if isinstance(el.text, unicode):
                        value['value'] = el.text
                    else:
                        value['value'] = el.text.decode('utf8')
                    datatype = el.attrib.get('datatype')
                    lang = el.attrib.get('lang')
                    if datatype:
                        value['datatype'] = datatype
                    elif lang:
                        value['lang'] = lang
                data[name] = value
            yield data
        



# Copyright (c) 2010 Infrae / Technical University Delft. All rights reserved.
# See also LICENSE.txt


from collections import defaultdict

from lxml import etree
from lxml.builder import ElementMaker

 
NAMESPACES = {'rdf': u'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
              'rdfs': u'http://www.w3.org/2000/01/rdf-schema#',
              'owl': u'http://www.w3.org/2002/07/owl#',
              'xsd': u'http://www.w3.org/2001/XMLSchema#',
              'fedora': u"info:fedora/fedora-system:def/relations-external#",
              'dc': u"http://purl.org/dc/elements/1.1/",
              'oai_dc': u"http://www.openarchives.org/OAI/2.0/oai_dc/",
              'dcterms': u'http://purl.org/dc/terms/'}

class Namespaces(dict):

    def __getattr__(self, prefix):
        value = self.get(prefix)
        if value is None:
            raise AttributeError('No such namespace prefix: %s' % prefix)

        ns = Namespace(value)
        return ns

    def url_split(self, url):
        name = url.rsplit('/', 1)[-1].rsplit(':', 1)[-1].rsplit('#', 1)[-1]
        ns = url[:-len(name)]
        return ns, name
        
    def prefix_url(self, url):
        ns, name = self.url_split(url)
        prefix = [prefix for (prefix, n) in self.items() if n == ns]
        if not prefix:
            raise ValueError('Can not prefix URL: "%s"' % url)
        return '%s:%s' % (prefix[0], name)

    def expand_url(self, url):
        prefix, rest = url.split(':', 1)
        try:
            ns = self[prefix]
        except KeyError:
            raise ValueError('Can not expand URL: "%s"' % url)
        return ns + rest

class Namespace(unicode):

    @property
    def title(self):
        return '%stitle' % self

    @property
    def format(self):
        return '%sformat' % self
    
    def __getattr__(self, name):
        return '%s%s' % (self, name)

    def __getitem__(self, name):
        return self.__getattr__(name)

NS = Namespaces(NAMESPACES)
NSXML = 'http://www.w3.org/XML/1998/namespace'

def dict2rdfxml(subject, predicates):
    rdf = ElementMaker(namespace=NS.rdf, nsmap=dict(NS))
    doc = rdf.RDF()
    subject = u'info:fedora/%s' % subject
    description= rdf.Description()
    description.attrib['{%s}about' % NS.rdf] = subject
    doc.append(description)
    
    for predicate, objects in predicates.items():
        ns, tagname = NS.url_split(predicate)
        for object in objects:
            el = etree.SubElement(description, '{%s}%s' % (ns, tagname))
            if object['type'] == 'uri':
                el.attrib['{%s}resource' % NS.rdf] = object['value']
            elif object['type'] == 'bnode':
                raise ValueError('Bnodes are not supported in Fedora')
            else:
                el.text = object['value']
                if object.get('lang'):
                    el.attrib['{%s}lang' % NSXML] = object['lang']
                elif object.get('datatype'):
                    el.attrib['{%s}datatype' % NS.rdf] = object['datatype']
                    
    return etree.tostring(doc,
                          encoding='UTF-8',
                          pretty_print=True,
                          xml_declaration=False)

def rdfxml2dict(rdfxml):
    doc = etree.fromstring(rdfxml)
    subject = None
    result = defaultdict(list)
    for description in doc:
        about = description.attrib['{%s}about' % NS.rdf]
        if subject is None:
            subject = about
        elif subject != about:
            raise ValueError('Statements about multiple subjects')
        for predicate in description:
            ns, tag = predicate.tag[1:].split('}')
            uri = predicate.attrib.get('{%s}resource' % NS.rdf)
            if uri:
                data = {'value': uri, 'type': 'uri'}
            else:
                data = {'value': predicate.text, 'type': 'literal'}
                lang = predicate.get('{%s}lang' % NSXML)
                datatype = predicate.get('{%s}datatype' % NS.rdf)
                if lang:
                    data['lang'] = lang
                elif datatype:
                    data['datatype'] = datatype
            result[ns+tag].append(data)
    return result



# Copyright (c) 2010 Infrae / Technical University Delft. All rights reserved.
# See also LICENSE.txt
import StringIO
import socket
import httplib
import urlparse
import logging

class APIException(Exception):
    """ An exception in the general usage of the API """
    pass

    
class FedoraConnectionException(Exception):
    """ An exception thrown by Fedora connections """
    def __init__(self, httpcode, reason=None, body=None):
        self.httpcode = httpcode
        self.reason = reason
        self.body = body

    def __repr__(self):
        return 'HTTP code=%s, Reason=%s, body=%s' % (
                    self.httpcode, self.reason, self.body.splitlines()[0])

    def __str__(self):
        return repr(self)


class Connection(object):
    """
    Represents a connection to a Fedora-Commons Repository using the REST API
    http://fedora-commons.org/confluence/display/FCR30/REST+API    
    """
    def __init__(self, url, debug=False,
                 username=None, password=None, 
                 persistent=True):
        """
         url -- URI pointing to the Fedora server. eg.
         
            http://localhost:8080/fedora/
            
         persistent -- Keep a persistent HTTP connection open.
                Defaults to true
        """        
        self.scheme, self.host, self.path = urlparse.urlparse(url, 'http')[:3]
        self.url = url
        self.username = username
        self.password = password
        self.debug = debug
        
        self.persistent = persistent
        self.reconnects = 0
        self.conn = httplib.HTTPConnection(self.host)

        self.form_headers = {}
        
        if not self.persistent:
            self.form_headers['Connection'] = 'close'
        
        if self.username and self.password:
            token = ('%s:%s' % (self.username,
                                self.password)).encode('base64').strip()
            self.form_headers['Authorization'] = 'Basic %s' % token
        
    def close(self):
        self.conn.close()

    def open(self, url, body='', headers=None, method='GET',):
        if headers is None:
            headers = {}
        if url.startswith('/'):
            url = url[1:]
        url = '%s/%s' % (self.path, url)
        attempts = 3

        while attempts:
            try:
                logging.debug('Trying %s on %s' % (method, url))
                self.conn.request(method, url, body, headers)
                return check_response_status(self.conn.getresponse())
            except (socket.error,
                    httplib.ImproperConnectionState,
                    httplib.BadStatusLine):
                    # We include BadStatusLine as they are spurious
                    # and may randomly happen on an otherwise fine
                    # connection (though not often)
                logging.exception('Got an Exception in open')
                self._reconnect()
                attempts -= 1
                if not attempts:
                    raise
        if not self.persistent:
           self.close()
        
    def _reconnect(self):
        self.reconnects += 1
        self.close()
        self.conn.connect()
        
def check_response_status(response):
    if response.status not in (200, 201, 204):
        ex = FedoraConnectionException(response.status, response.reason)
        try:
            ex.body = response.read()
        except:
            pass
        raise ex
    return response



from setuptools import setup, find_packages
from os.path import join, dirname

HISTORY = join(dirname(__file__), 'HISTORY.txt')
README = join(dirname(__file__), 'src', 'fcrepo', 'README.txt')

setup(
    name='fcrepo',
    version='1.0b1',
    author='Infrae / Jasper Op de Coul',
    author_email='jasper@infrae.com',
    description="API implementation for the Fedora Commons Repository platform",
    long_description=(open(README).read()+
                      '\n'+
                      open(HISTORY).read()),
    classifiers=["Development Status :: 4 - Beta",
                 "Programming Language :: Python",
                 "License :: OSI Approved :: BSD License",
                 "Topic :: Software Development :: Libraries :: Python Modules",
                 "Environment :: Web Environment",
                 "Intended Audience :: Science/Research"],

    packages=find_packages('src'),
    package_dir = {'': 'src'},
    include_package_data = True,
    zip_safe=False,
    license='BSD',
    entry_points= {
    'console_scripts': [
        'install_fedora = fcrepo.scripts:install_fedora',
        'start_fedora = fcrepo.scripts:start_fedora',
      ]
    },
    install_requires=[
    'lxml',
    ],
)



# Copyright (c) 2010 Infrae / Technical University Delft. All rights reserved.
# See also LICENSE.txt

from collections import defaultdict

from lxml import etree

from fcrepo.utils import rdfxml2dict, dict2rdfxml
    
class typedproperty(property):
    def __init__(self, fget, fset=None, fdel=None, doc=None, pytype=None):
        # like a normal property, but converts types to/from strings
        def typed_get(self):
            if pytype is bool:
                return fget(self) == 'true'
            return pytype(fget(self))
        
        def typed_set(self, value):
            if pytype is bool:
                return fset(self, unicode(value).lower())
            else:
                return fset(self, unicode(value).lower())
            
        super(typedproperty, self).__init__(typed_get, typed_set, fdel, doc)

class FedoraDatastream(object):
    def __init__(self, dsid, object):
        self.object = object
        self.dsid = dsid
        self._info = self.object.client.getDatastreamProfile(self.object.pid,
                                                             self.dsid)

        
    def delete(self, **params):
        self.object.client.deleteDatastream(self.object.pid,
                                            self.dsid,
                                            **params)
        self.object._dsids = None

    def getContent(self):
        return self.object.client.getDatastream(self.object.pid, self.dsid)

    def setContent(self, data='', **params):
            
        if self._info['controlGroup'] == 'X':
            # for some reason we need to add 2 characters to the body
            # or we get a parsing error in fedora
            data += '\r\n'
        
        self.object.client.modifyDatastream(self.object.pid,
                                            self.dsid,
                                            data,
                                            **params)
        self._info = self.object.client.getDatastreamProfile(self.object.pid,
                                                             self.dsid)
        
    def _setProperty(self, name, value):
        msg = u'Changed %s datastream property' % name
        name = {'label': 'dsLabel',
                'location': 'dsLocation',
                'state': 'dsState'}.get(name, name)
        params = {name: value, 'logMessage': msg, 'ignoreContent': True}
        self.object.client.modifyDatastream(self.object.pid,
                                            self.dsid,
                                            **params)
        self._info = self.object.client.getDatastreamProfile(self.object.pid,
                                                             self.dsid)

    label = property(lambda self: self._info['label'],
                     lambda self, value: self._setProperty('label', value))
    location = property(lambda self: self._info['location'],
                        lambda self, value: self._setProperty('location', value))
    state = property(lambda self: self._info['state'],
                           lambda self, value: self._setProperty('state',
                                                                 value))
    checksumType = property(lambda self: self._info['checksumType'],
                            lambda self, value: self._setProperty('checksumType',
                                                                 value))
    versionId = property(lambda self: self._info['versionId'],
                        lambda self, value: self._setProperty('versionId',
                                                              value)) 
    mimeType = property(lambda self: self._info['mimeType'],
                        lambda self, value: self._setProperty('mimeType',
                                                              value)) 
    formatURI = property(lambda self: self._info['formatURI'],
                         lambda self, value: self._setProperty('formatURI',
                                                               value)) 


    versionable = typedproperty(lambda self: self._info['versionable'],
                                lambda self, value: self._setProperty(
                                  'versionable', value), pytype=bool) 

    # read only
    createdDate = property(lambda self: self._info['createdDate'])
    controlGroup = property(lambda self: self._info['controlGroup'])
    size = typedproperty(lambda self: self._info['size'], pytype=int)
    checksum = property(lambda self: self._info['formatURI'])



class RELSEXTDatastream(FedoraDatastream):
    def __init__(self, dsid, object):
        super(RELSEXTDatastream, self).__init__(dsid, object)
        self._rdf = None

    def _get_rdf(self):
        if self._rdf is None:
            rdfxml = self.getContent().read()
            self._rdf = rdfxml2dict(rdfxml)
        return self._rdf

    def keys(self):
        rdf = self._get_rdf()
        keys = rdf.keys()
        keys.sort()
        return keys
    predicates = keys
    
    def setContent(self, data='', **params):
        if not data:
            rdf = self._get_rdf()
            data = dict2rdfxml(self.object.pid, rdf)
            self._rdf = None
        super(RELSEXTDatastream, self).setContent(data, **params)
    def __setitem__(self, key, value):
        rdf = self._get_rdf()
        rdf[key]=value
        
    def __getitem__(self, key):
        rdf = self._get_rdf()
        return rdf[key]
    
    def __delitem__(self, key):
        rdf = self._get_rdf()
        del rdf[key]

    def __contains__(self, key):
        rdf = self._get_rdf()
        return key in rdf

    def __iter__(self):
        rdf = self._get_rdf()
        return rdf.__iter__()
    
class DCDatastream(FedoraDatastream):
    def __init__(self, dsid, object):
        super(DCDatastream, self).__init__(dsid, object)
        self._dc = None

    def _get_dc(self):
        if self._dc is None:
            xml = self.getContent().read()
            doc = etree.fromstring(xml)
            self._dc = defaultdict(list)
            for child in doc:
                name = child.tag.split('}')[-1]
                value = child.text
                if not isinstance(value, unicode):
                    value = value.decode('utf8')
                self._dc[name].append(value)
        return self._dc

    def keys(self):
        dc = self._get_dc()
        keys = dc.keys()
        keys.sort()
        return keys
    properties = keys
    
    def setContent(self, data='', **params):
        if not data:
            dc = self._get_dc()
            nsmap = {'dc': 'http://purl.org/dc/elements/1.1/',
                     'oai_dc': 'http://www.openarchives.org/OAI/2.0/oai_dc/'}
            doc = etree.Element('{%s}dc' % nsmap['oai_dc'], nsmap=nsmap)
            for key, values in dc.items():
                for value in values:
                    el = etree.SubElement(doc, '{%s}%s' % (nsmap['dc'], key))
                    el.text = value
            data = etree.tostring(doc, encoding="UTF-8",
                                  pretty_print=True, xml_declaration=False)
            self._dc = None
        super(DCDatastream, self).setContent(data, **params)
        
    def __setitem__(self, key, value):
        dc = self._get_dc()
        dc[key]=value
        
    def __getitem__(self, key):
        dc = self._get_dc()
        return dc[key]
    
    def __delitem__(self, key):
        dc = self._get_dc()
        del dc[key]

    def __contains__(self, key):
        dc = self._get_dc()
        return key in dc

    def __iter__(self):
        dc = self._get_dc()
        return dc.__iter__()
